use std::{
    fs::File,
    io::{Cursor, Read},
};

use ruzstd::StreamingDecoder;
use vibrato::{Dictionary, Tokenizer};

const DICT: &[u8] = include_bytes!("../dictionary/system.dic.zst");

fn main() {
    // Loads a compiled dictionary
    let dict = {
        let decoder = StreamingDecoder::new(Cursor::new(DICT)).unwrap();
        Dictionary::read(decoder).unwrap()
    };

    let text = {
        let mut text = String::new();
        File::open(std::env::args().nth(1).unwrap())
            .unwrap()
            .read_to_string(&mut text)
            .unwrap();
        text
    };

    let tokenizer = Tokenizer::new(dict);
    println!("{}", add_furigana(&text, &tokenizer));
}

fn add_furigana(text: &str, tokenizer: &Tokenizer) -> String {
    let mut worker = tokenizer.new_worker();
    worker.reset_sentence(text);
    worker.tokenize();

    let mut new_text = String::new();
    for i in 0..worker.num_tokens() {
        let t = worker.token(i);
        let surface = t.surface();
        let kana = t.feature().split(",").nth(1).unwrap();

        let (start_bytes, end_bytes) = matching_ends(surface, kana);

        if kana.is_empty()
            || start_bytes == surface.len()
            || surface
                .chars()
                .map(|c| c.is_ascii() || c.is_numeric())
                .all(|n| n)
        {
            new_text.push_str(surface);
        } else {
            let start = &surface[..start_bytes];
            let mid = &surface[start_bytes..(surface.len() - end_bytes)];
            let mid_kana = &kana[start_bytes..(kana.len() - end_bytes)];
            let end = &surface[(surface.len() - end_bytes)..];
            new_text.push_str(start);
            new_text.push_str(" ");
            new_text.push_str(mid);
            new_text.push_str("[");
            new_text.push_str(mid_kana);
            new_text.push_str("]");
            new_text.push_str(end);
        }
    }

    new_text
}

/// Returns (matching_start_bytes, matching_end_bytes).
fn matching_ends(a: &str, b: &str) -> (usize, usize) {
    let mut start_bytes = 0;
    for (ca, cb) in a.chars().zip(b.chars()) {
        if ca == cb || is_equivalent_kana(ca, cb) {
            start_bytes += ca.len_utf8();
        } else {
            break;
        }
    }

    let mut end_bytes = 0;
    for (ca, cb) in a.chars().rev().zip(b.chars().rev()) {
        if ca == cb || is_equivalent_kana(ca, cb) {
            end_bytes += ca.len_utf8();
        } else {
            break;
        }
    }

    if start_bytes == a.len() {
        (start_bytes, 0)
    } else {
        (start_bytes, end_bytes)
    }
}

fn is_equivalent_kana(a: char, b: char) -> bool {
    let a = normalize_kana(a);
    let b = normalize_kana(b);
    match (a, b) {
        (Some('は'), Some('わ'))
        | (Some('わ'), Some('は'))
        | (Some('を'), Some('お'))
        | (Some('お'), Some('を'))
        | (Some(_), Some('ー'))
        | (Some('ー'), Some(_)) => true,

        (Some(c), Some(d)) if c == d => true,

        _ => false,
    }
}

const HIRAGANA: u32 = 0x3041;
const KATAKANA: u32 = 0x30A1;
const KANA_COUNT: u32 = 0x3097 - HIRAGANA;

pub fn is_kana(c: char) -> bool {
    if c == 'ー' {
        return true;
    }

    let c = c as u32;

    if c >= HIRAGANA && c < (HIRAGANA + KANA_COUNT) {
        return true;
    }

    if c >= KATAKANA && c < (KATAKANA + KANA_COUNT) {
        return true;
    }

    return false;
}

pub fn normalize_kana(c: char) -> Option<char> {
    if !is_kana(c) {
        return None;
    }

    Some(katakana_to_hiragana(c).unwrap_or(c))
}

pub fn hiragana_to_katakana(c: char) -> Option<char> {
    let c = c as u32;
    if c >= HIRAGANA && c < (HIRAGANA + KANA_COUNT) {
        char::try_from(c + KATAKANA - HIRAGANA).ok()
    } else {
        None
    }
}

pub fn katakana_to_hiragana(c: char) -> Option<char> {
    let c = c as u32;
    if c >= KATAKANA && c < (KATAKANA + KANA_COUNT) {
        char::try_from(c - KATAKANA + HIRAGANA).ok()
    } else {
        None
    }
}
